{
    VB style Property Editor Clone Component for Borland Delphi 2

    Author         : Wong Ping Yung (William)
    Version        : 1.1
    First Created  : 12th October 1996 (very alpha)
    Last Modified  : 2nd January, 1997.
    First Released : 2nd Jaunary, 1997.
}

Unit TPEComp;

interface

uses
    Windows, SysUtils, Classes, Controls, StdCtrls, Buttons, Grids, Dialogs,
    Messages, DsgnIntf, TPEAbout, TPERef, TNumEdit, Spin;

const
    emDropDownList   = 0;
    emUserEdit       = 1;
    emUserBrowse     = 2;
    emUserEditBrowse = 3;
    emFileBrowse     = 4;
    emFontBrowse     = 5;
    emColorBrowse    = 6;

type
{Component Editor}
    TCompEditor = class(TComponentEditor)
    private
        procedure ExecuteVerb(Index: integer); override;
        function  GetVerb(Index: integer): string; override;
        function  GetVerbCount: integer; override;
    end;
{Component itself}
    TUserBrowser = function(var Value: variant): boolean;
    PProperty = ^TProperty;
    TProperty = record
        Name        : string;       // Property name
        EditMode    : integer;      // Edit mode for property
        Filter      : string;       // Filter used in File Open Dialog if any
        MaxValue    : variant;      // Maximum value allowed
        MinValue    : variant;      // Minimum value allowed
        MaxLength   : word;         // Maximum length of input, 0 for none
        NDigits     : byte;         // Maximum no of digits in floating point
        Visible     : boolean;      // Visiblity of property
        DefVisible  : boolean;      // User prefered Visiblity
        Value       : variant;      // Actual value of property
        ValueSet    : variant;      // Set of possible values for property if any
        UserBrowser : TUserBrowser; // Custom browser if any
        DependOn    : string;       // Visible depends on other property
        DependValue : variant;      // Value to check 
    end;
    PObj = ^TObj;
    TObj = record
        Name  : string; // Name of object page
        Props : TList;  // Properties in object
    end;
    TPropertiesEdit = class(TCustomControl)
    private
        { Private Declarations }
        FObjectNames: TComboBox;         // Combobox for object selection
        FPropPullDown: TComboBox;        // Combobox for emDropDownList
        FPropStringEdit: TEdit;          // Editor for property of string type
        FPropFloatEdit: TFloatEdit;      // Editor for property of float type
        FPropIntegerEdit: TIntegerEdit;  // Editor for property of integer type
        FPropIntegerUpDown: TSpinButton; // Updown arrows for integer editor
        FOKBtn: TSpeedButton;            // OK button
        FCancelBtn: TSpeedButton;        // Cancel button
        FBrowseBtn: TSpeedButton;        // Browse button
        FTable: TStringGrid;             // Table to display/select property
        FOpenDialog: TOpenDialog;        // File open dialog
        FFontDialog: TFontDialog;        // Font dialog
        FColorDialog: TColorDialog;      // Color dialog
        FObjects: TList;                 // Object List
        FAutoHide: boolean;              // Enable/disable auto-hide of property
        FEnabled: boolean;               // Disable/Enable editor
        FNumObjects: integer;            // Total number of objects being edited
        FPropertyIndex: integer;         // Index for selected property
        FObjectIndex: integer;           // Index for selected object
        FSorted: boolean;                // Table of property is sorted/unsorted
        FShowObjectList: boolean;        // Display/Hide combobox of objects
        FOnPropertyChange: TNotifyEvent; // Event occurs when property changed
        FOnPropertyValueChange: TNotifyEvent; // Event occurs when value of
                                              // property changed
        FOnPropertyAfterEdit: TNotifyEvent;   // Event occurs after user edits
                                              // any property
    protected
        { Protected Declarations }
        procedure Paint; override;
        procedure SetPropertyIndex(Value: integer);
        procedure SetObjectIndex(Value: integer);
        procedure SetShowObjectList(Value: boolean);
        procedure SetEnabled(Value: boolean);
        procedure FTableSelect(Sender: TObject; Col,Row: longint;
                               var CanSelect: boolean);
        procedure FObjectChange(Sender: TObject);
        procedure FPropPullDownValuesChange(Sender: TObject);
        procedure FPropEditKeyPress(Sender: TObject; var Key: char);
        procedure FPropEditChange(Sender: TObject);
        procedure FOKBtnClick(Sender: TObject);
        procedure FCancelBtnClick(Sender: TObject);
        procedure FBrowseBtnClick(Sender: TObject);
        procedure SetDefaultProp(Prop: PProperty);
        function  GetValue(index: integer) : variant;
        procedure SetValue(index: integer; Value: variant);
        function  GetMaxValue(index: integer) : variant;
        procedure SetMaxValue(index: integer; Value: variant);
        function  GetMinValue(index: integer) : variant;
        procedure SetMinValue(index: integer; Value: variant);
        function  GetMaxLength(index: integer) : word;
        procedure SetMaxLength(index: integer; Value: word);
        function  GetColWidth(index: integer): integer;
        procedure SetColWidth(index: integer; Value: integer);
        function  GetNumProps(ObjIndex: integer): integer;
        function  GetPropertyName(index: integer): string;
        procedure SetPropertyName(index: integer; PropName: string);
        function  GetPropertyFilter(index: integer): string;
        procedure SetPropertyFilter(index: integer; Filter: string);
        function  GetValueSet(index: integer): variant;
        procedure SetValueSet(index: integer; ValueSet: variant);
        function  GetEditMode(index: integer): integer;
        procedure SetEditMode(index,EditMode: integer);
        function  GetPropertyBrowser(index: integer): TUserBrowser;
        procedure SetPropertyBrowser(index: integer; Browser: TUserBrowser);
        procedure SetSorted(Value: boolean);
        procedure SpinUp(Sender: TObject);
        procedure SpinDown(Sender: TObject);
        function  GetVisible(index: integer): boolean;
        procedure SetVisible(index: integer; Value: boolean);
        function  GetNDigits(index: integer): byte;
        procedure SetNDigits(index: integer; Value: byte);
        procedure SetAutoHide(Value: boolean);
        function  GetDependOn(index: integer): string;
        procedure SetDependOn(index: integer; Value: string);
        function  GetDependValue(index: integer): variant;
        procedure SetDependValue(index: integer; Value: variant);
        procedure FAfterEdit(Sender: TObject);
    public
        { Public Declarations }
        constructor Create(AOwner: TComponent); override;
        destructor  Destroy; override;
        function    AddObject(Name: string) : integer;
        function    AddPropertyVar(Name: string; EditMode: integer;
                                   ValueSet: variant; Value: variant): integer;
        function    AddPropertyConst(Name: string; EditMode: integer;
                                     const ValueSet: array of variant;
                                     Value: variant): integer;
        property    AutoHide: boolean read FAutoHide write SetAutoHide;
        property    Browser[index: integer]: TUserBrowser
                    read GetPropertyBrowser write SetPropertyBrowser;
        procedure   Clear;
        property    ColWidths[index: integer]: integer
                    read GetColWidth write SetColWidth;
        procedure   DeleteObject(index: integer);
        procedure   DeleteProperty(index: integer);
        property    DependOn[index: integer]: string
                    read GetDependOn write SetDependOn;
        property    DependValue[index: integer]: variant
                    read GetDependValue write SetDependValue;
        property    EditMode[index: integer]: integer
                    read GetEditMode write SetEditMode;
        procedure   ExchangeObject(index1, index2: integer);                    
        procedure   ExchangeProperty(index1, index2: integer);
        property    Filter[index: integer]: string
                    read GetPropertyFilter write SetPropertyFilter;
        function    InsertObject(index: integer; Name: string) : integer;
        function    InsertPropertyVar(index: integer; Name: string; EditMode: integer;
                                      ValueSet: variant; Value: variant): integer;
        function    InsertPropertyConst(index: integer; Name: string; EditMode: integer;
                                        const ValueSet: array of variant;
                                        Value: variant): integer;
        property    MaxLength[index: integer]: word
                    read GetMaxLength write SetMaxLength;
        property    MaxValue[index: integer]: variant
                    read GetMaxValue write SetMaxValue;
        property    MinValue[index: integer]: variant
                    read GetMinValue write SetMinValue;
        property    NDigits[index: integer]: byte read GetNDigits write SetNDigits;
        property    NumProps[ObjIndex: integer]: integer read GetNumProps;
        function    ObjectIndexOf(Name: string): integer;
        function    PropertyIndexOf(Name: string): integer;
        property    PropertyName[index: integer]: string
                    read GetPropertyName write SetPropertyName;
        property    Value[index: integer]: variant
                    read GetValue write SetValue;
        property    ValueSet[index: integer]: variant
                    read GetValueSet write SetValueSet;
        property    IsVisible[index: integer]: boolean
                    read GetVisible write SetVisible;
    published
        property Align;
        property ClientHeight;
        property ClientWidth;
        property Ctl3D;
        property Enabled: boolean read FEnabled write SetEnabled;
        property Font;
        property Hint;
        property NumObjs: integer read FNumObjects;
        property ObjectIndex: integer
                    read FObjectIndex write SetObjectIndex;
        property PopupMenu;
        property PropertyIndex: integer
                    read FPropertyIndex write SetPropertyIndex;
        property ShowObjectList: boolean
                    read FShowObjectList write SetShowObjectList;
        property ParentCtl3D;
        property ParentFont;
        property ParentSHowHint;
        property ShowHint;
        property Sorted: boolean read FSorted write SetSorted;
        property TabStop;
        property TabOrder;
        property Tag;
        property Visible;
        property OnEnter;
        property OnExit;
        property OnPropertyChange: TNotifyEvent
                    read FOnPropertyChange write FOnPropertyChange;
        property OnPropertyValueChange: TNotifyEvent
                    read FOnPropertyValueChange write FOnPropertyValueChange;
        property OnPropertyAfterEdit: TNotifyEvent
                    read FOnPropertyAfterEdit write FOnPropertyAfterEdit;
    end;

procedure Register;

implementation

{$R TPEComp.RES}

function CompareProp(Prop1,Prop2: pointer): integer;
begin
    if PProperty(Prop1)^.Name<PProperty(Prop2)^.Name then
        Result := -1
    else if PProperty(Prop1)^.Name=PProperty(Prop2)^.Name then
        Result := 0
    else
        Result := 1;
end;

{Implementation of the component}
constructor TPropertiesEdit.Create(AOwner: TComponent);
begin
    inherited Create(AOwner);
    ControlStyle := ControlStyle + [csOpaque];
    Width := 200;
    Height := 200;
    FObjects := TList.Create;
    FNumObjects := 0;
    FPropertyIndex := -1;
    FObjectIndex := -1;
    FShowObjectList := True;
    FEnabled := True;
    FSorted := False;
    FAutoHide := True;
{Create necessary components}
    FObjectNames := TComboBox.Create(self);
    FBrowseBtn := TSpeedButton.Create(self);
    FCancelBtn := TSpeedButton.Create(self);
    FOKBtn := TSpeedButton.Create(self);
    FPropIntegerUpDown := TSpinButton.Create(self);
    FPropStringEdit := TEdit.Create(self);
    FPropFloatEdit := TFloatEdit.Create(self);
    FPropIntegerEdit := TIntegerEdit.Create(self);
    FPropPullDown := TComboBox.Create(self);
    FTable := TStringGrid.Create(self);
{Adjust dropdown list component for object name}
    with FObjectNames do begin
        Parent := self;
        Align := alTop;
        Style := csDropDownList;
        Visible := True;
        Enabled := True;
        OnChange := FObjectChange;
    end;
{Adjust Cancel button}
    with FCancelBtn do begin
        Parent := self;
        Align := alNone;
        SetBounds(FObjectNames.Left,FObjectNames.Top+FObjectNames.Height,20,20);
        Visible := True;
        Enabled := True;
        Glyph.Handle := LoadBitmap(HInstance,'CANCELBUTTON');
        Glyph.Height := 16;
        Glyph.Width := 16;
        OnClick := FCancelBtnClick;
    end;
{Adjust OK button}
    with FOKBtn do begin
        Parent := self;
        Align := alNone;
        SetBounds(FCancelBtn.Left+FCancelBtn.Width,FCancelBtn.Top,
                  FCancelBtn.Width,FCancelBtn.Height);
        Visible := True;
        Enabled := True;
        Glyph.Handle := LoadBitmap(HInstance,'OKBUTTON');
        Glyph.Height := 16;
        Glyph.Width := 16;
        OnClick := FOKBtnClick;
    end;
{Adjust edit component for property value}
    with FPropStringEdit do begin
        Parent := self;
        Align := alNone;
        Visible := False;
        Enabled := True;
        OnKeyPress := FPropEditKeyPress;
        OnExit := FAfterEdit;
    end;
    with FPropFloatEdit do begin
        Parent := self;
        Align := alNone;
        Visible := False;
        Enabled := True;
        Digits := 18;
        OnKeyPress := FPropEditKeyPress;
        OnExit := FAfterEdit;
    end;
    with FPropIntegerEdit do begin
        Parent := self;
        Align := alNone;
        Visible := False;
        Enabled := True;
        OnKeyPress := FPropEditKeyPress;
        OnExit := FAfterEdit;
    end;
    with FPropIntegerUpDown do begin
        Parent := self;
        Visible := False;
        Enabled := True;
        Align := alNone;
        OnUpClick := SpinUp;
        OnDownClick := SpinDown;
    end;
{Adjust dropdown list component for property values}
    with FPropPullDown do begin
        Parent := self;
        Align := alNone;
        Style := csDropDownList;
        Visible := True;
        Enabled := True;
        OnChange := FPropPullDownValuesChange;
    end;
{Adjust Link button}
    with FBrowseBtn do begin
        Parent := self;
        Align := alNone;
        SetBounds(self.Width-FBrowseBtn.Width,FOKBtn.Top,
                  FOKBtn.Width,FOKBtn.Width);
        Visible := False;
        Enabled := True;
        Glyph.Handle := LoadBitmap(HInstance,'BROWSEBUTTON');
        Glyph.Height := 16;
        Glyph.Width := 16;
        OnClick := FBrowseBtnClick;
    end;
{Adjust table}
    with FTable do begin
        Parent := self;
        Align := alNone;
        Options := [goFixedHorzLine,goVertLine,goHorzLine,
                    goColSizing,goRowSelect];
        ColCount := 2;
        RowCount := 1;
        FixedCols := 0;
        FixedRows := 0;
        Cells[0,0] := 'Name';
        Cells[1,0] := 'Value';
        Visible := True;
        Enabled := True;
        OnSelectCell := FTableSelect;
        OnDblClick := FBrowseBtnClick;
    end;
{Create File Open dilaog component}
    FOpenDialog := TOpenDialog.Create(self);
    with FOpenDialog do begin
        Options := [ofFileMustExist,ofHideReadOnly,ofShareAware];
    end;
{Create Font Dialog component}
    FFontDialog := TFontDialog.Create(self);
    with FFontDialog do begin
        Options := [fdForceFontExist];
    end;
{Create Color Dialog component}
    FColorDialog := TColorDialog.Create(self);
    with FColorDialog do begin
        Options := [cdFullOpen];
    end;
end;

destructor TPropertiesEdit.Destroy;
var
    i,j : integer;
    Obj : PObj;
begin
    for i := 0 to FObjects.Count-1 do begin
        Obj := FObjects.Items[i];
        for j := 0 to Obj^.Props.Count-1 do dispose(Obj^.Props.Items[j]);
        Obj^.Props.Free;
        dispose(Obj);
    end;
    FObjects.Free;
    FObjectNames.Free;
    FPropPullDown.Free;
    FPropStringEdit.Free;
    FPropFloatEdit.Free;
    FPropIntegerEdit.Free;
    FPropIntegerUpDown.Free;
    FOKBtn.Free;
    FCancelBtn.Free;
    FBrowseBtn.Free;
    FTable.Free;
    FOpenDialog.Free;
    FFontDialog.Free;
    FColorDialog.Free;
    inherited Destroy;
end;

procedure TPropertiesEdit.FAfterEdit(Sender: TObject);
begin
    if Assigned(FOnPropertyAfterEdit) then FOnPropertyAfterEdit(Sender);
end;

procedure TPropertiesEdit.Clear;
begin
    FObjects.Clear;
    FObjectIndex := -1;
    FPropertyIndex := -1;
    if Visible then Paint;
end;

procedure TPropertiesEdit.SetObjectIndex(Value: integer);
begin
    FObjectIndex := Value;
    if Visible then Paint;
end;

procedure TPropertiesEdit.SetPropertyIndex(Value: integer);
begin
    if (FObjectIndex<0) or (FObjectIndex>=FObjects.Count) then exit;
    if (Value<PObj(FObjects.Items[FObjectIndex])^.Props.Count) then
        FPropertyIndex := Value
    else
        FPropertyIndex := PObj(FObjects.Items[FObjectIndex])^.Props.Count-1;
    if Visible then Paint;
end;

procedure TPropertiesEdit.SetShowObjectList(Value: boolean);
begin
    FShowObjectList := Value;
    if Visible then Paint;
end;

function TPropertiesEdit.AddObject(Name: string) : integer;
begin
    Result := InsertObject(FObjects.Count+1,Name);
end;

function TPropertiesEdit.InsertObject(index: integer; Name: string) : integer;
var
    Obj : PObj;
begin
    new(Obj);
    Obj^.Name := Name;
    Obj^.Props := TList.Create;
    if index<FObjects.Count then begin
        FObjectNames.Items.Insert(index,Name);
        FObjects.Insert(index,Obj); end
    else begin
        FObjectNames.Items.Add(Name);
        FObjects.Add(Obj);
    end;
    FNumObjects := FObjects.Count;
    Result := FObjects.IndexOf(Obj);
    if Visible then Paint;
end;

function TPropertiesEdit.AddPropertyVar(Name: string; EditMode: integer;
            ValueSet: variant; Value: variant): integer;
begin
    Result := InsertPropertyVar(PObj(FObjects.Items[FObjectIndex])^.Props.Count+1,
                                  Name,EditMode,ValueSet,Value);
end;

function TPropertiesEdit.InsertPropertyVar(index: integer; Name: string; EditMode: integer;
            ValueSet: variant; Value: variant): integer;
var
    Obj     : PObj;
    Prop    : PProperty;
    i,lo,hi : integer;
begin
    new(Prop);
    Prop^.Name := Name;
    Prop^.EditMode := EditMode;
    Prop^.Value := Value;
    if VarIsArray(ValueSet) then begin
        lo := VarArrayLowBound(ValueSet,1);
        hi := VarArrayHighBound(ValueSet,1);
        if (lo<=hi) then begin
            Prop^.ValueSet := VarArrayCreate([lo,hi],VarVariant);
            for i := lo to hi do
                Prop^.ValueSet[i] := ValueSet[i];
        end; end
    else begin
        VarClear(Prop^.ValueSet);
    end;
    SetDefaultProp(Prop);
    Obj := FObjects.Items[FObjectIndex];
    if index<Obj^.Props.Count then
        Obj^.Props.Insert(index,Prop)
    else
        Obj^.Props.Add(Prop);
    if FSorted then Obj^.Props.Sort(CompareProp);
    Result := Obj^.Props.IndexOf(Prop);
    if Visible then Paint;
end;

procedure TPropertiesEdit.SetDefaultProp(Prop: PProperty);
begin
    Prop^.MaxLength := 0;
    Prop^.Visible := True;
    Prop^.DefVisible := True;
    Prop^.NDigits := 18;
    Prop^.Filter := '';
    Prop^.UserBrowser := nil;
    Prop^.DependOn := '';
    VarClear(Prop^.DependValue);
    VarClear(Prop^.MaxValue);
    VarClear(Prop^.MinValue);
    if Prop^.EditMode in [emUserEdit,emUserEditBrowse] then begin
        case VarType(Prop^.Value) of
            varInteger : begin
                Prop^.MaxValue := High(Integer);
                Prop^.MinValue := Low(Integer); end;
            varSmallInt : begin
                Prop^.MaxValue := High(SmallInt);
                Prop^.MinValue := Low(SmallInt); end;
            varByte : begin
                Prop^.MaxValue := High(Byte);
                Prop^.MinValue := Low(Byte); end;
            varSingle : begin
                Prop^.MaxValue := 3.4e38;
                Prop^.MinValue := -3.4e38; end;
            varDouble : begin
                Prop^.MaxValue := 1.7e308;
                Prop^.MinValue := -1.7e308; end;
            varCurrency : begin
                Prop^.MaxValue := 922337203685477.5807;
                Prop^.MinValue := -922337203685477.5807;
            end;
        end;
    end;
end;

function TPropertiesEdit.AddPropertyConst(Name: string; EditMode: integer;
            const ValueSet: array of variant; Value: variant): integer;
begin
    Result := InsertPropertyConst(PObj(FObjects.Items[FObjectIndex])^.Props.Count+1,
                                  Name,EditMode,ValueSet,Value);
end;

function TPropertiesEdit.InsertPropertyConst(index: integer; Name: string; EditMode: integer;
            const ValueSet: array of variant; Value: variant): integer;
var
    Obj     : PObj;
    Prop    : PProperty;
    i,lo,hi : integer;
begin
    new(Prop);
    Prop^.Name := Name;
    Prop^.EditMode := EditMode;
    Prop^.Value := Value;
    lo := Low(ValueSet);
    hi := High(ValueSet);
    if (lo<=hi) and (not VarIsNull(ValueSet[lo])) then begin
        Prop^.ValueSet := VarArrayCreate([lo,hi],VarVariant);
        for i := lo to hi do
            Prop^.ValueSet[i] := ValueSet[i];
    end;
    SetDefaultProp(Prop);
    Obj := FObjects.Items[FObjectIndex];
    if index<Obj^.Props.Count then
        Obj^.Props.Insert(index,Prop)
    else
        Obj^.Props.Add(Prop);
    if FSorted then Obj^.Props.Sort(CompareProp);
    Result := Obj^.Props.IndexOf(Prop);
    if Visible then Paint;
end;

procedure TPropertiesEdit.DeleteObject(index: integer);
var
    Obj : PObj;
    i   : integer;
begin
    if (index<FNumObjects) and (index>=0) then begin
        Obj := FObjects.Items[index];
        for i := 0 to Obj^.Props.Count-1 do dispose(Obj^.Props.Items[i]);
        Obj^.Props.Free;
        dispose(Obj);
        FObjects.Delete(index);
        dec(FNumObjects);
    end;
    if Visible then Paint;
end;

procedure TPropertiesEdit.DeleteProperty(index: integer);
var
    Obj : PObj;
begin
    Obj := FObjects.Items[FObjectIndex];
    if (index<Obj^.Props.Count) and (index>=0) then begin
        dispose(Obj^.Props.Items[index]);
        Obj^.Props.Delete(index);
    end;
    if Visible then Paint;
end;

function TPropertiesEdit.GetValue(index: integer): variant;
var
    Obj  : PObj;
    Prop : PProperty;
begin
    Obj := FObjects.Items[FObjectIndex];
    Prop := Obj^.Props.Items[index];
    Result := Prop^.Value;
end;

procedure TPropertiesEdit.SetValue(index: integer; Value: variant);
var
    Obj : PObj;
    Prop : PProperty;
    tmp : variant;
begin
    Obj := FObjects.Items[FObjectIndex];
    Prop := Obj^.Props.Items[index];
    tmp := Prop^.Value;
    VarClear(Prop^.Value); // make sure no invalid variant type conversion
    Prop^.Value := Value;
    if Visible then Paint;
    if Assigned(FOnPropertyValueChange) then FOnPropertyValueChange(self);
    if Assigned(FOnPropertyValueChange) then
        if (VarType(tmp)<>VarType(Value)) then
            FOnPropertyValueChange(self)
        else if (tmp<>Value) then
            FOnPropertyValueChange(self);
end;

function TPropertiesEdit.GetMaxValue(index: integer): variant;
var
    Obj  : PObj;
    Prop : PProperty;
begin
    Obj := FObjects.Items[FObjectIndex];
    Prop := Obj^.Props.Items[index];
    Result := Prop^.MaxValue;
end;

procedure TPropertiesEdit.SetMaxLength(index: integer; Value: word);
var
    Obj : PObj;
    Prop : PProperty;
    tmp : variant;
begin
    Obj := FObjects.Items[FObjectIndex];
    Prop := Obj^.Props.Items[index];
    tmp := Prop^.MaxLength;
    Prop^.MaxLength := Value;
    if Visible then Paint;
    if (tmp<>Value) and Assigned(FOnPropertyChange) then
        FOnPropertyChange(self);
end;

function TPropertiesEdit.GetMaxLength(index: integer): word;
var
    Obj  : PObj;
    Prop : PProperty;
begin
    Obj := FObjects.Items[FObjectIndex];
    Prop := Obj^.Props.Items[index];
    Result := Prop^.MaxLength;
end;

procedure TPropertiesEdit.SetMaxValue(index: integer; Value: variant);
var
    Obj  : PObj;
    Prop : PProperty;
    tmp  : variant;
begin
    Obj := FObjects.Items[FObjectIndex];
    Prop := Obj^.Props.Items[index];
    tmp := Prop^.MaxValue;
    Prop^.MaxValue := Value;
    if Visible then Paint;
    if Assigned(FOnPropertyChange) then
        if (VarType(tmp)<>VarType(Value)) or (tmp<>Value) then
            FOnPropertyChange(self);
end;

function TPropertiesEdit.GetMinValue(index: integer): variant;
var
    Obj  : PObj;
    Prop : PProperty;
begin
    Obj := FObjects.Items[FObjectIndex];
    Prop := Obj^.Props.Items[index];
    Result := Prop^.MinValue;
end;

procedure TPropertiesEdit.SetMinValue(index: integer; Value: variant);
var
    Obj  : PObj;
    Prop : PProperty;
    tmp  : variant;
begin
    Obj := FObjects.Items[FObjectIndex];
    Prop := Obj^.Props.Items[index];
    tmp := Prop^.MinValue;
    Prop^.MinValue := Value;
    if Visible then Paint;
    if Assigned(FOnPropertyChange) then
        if (VarType(tmp)<>VarType(Value)) or (tmp<>Value) then
            FOnPropertyChange(self);
end;

function TPropertiesEdit.GetColWidth(index: integer): integer;
begin
    if not index in [0,1] then
        Result := 0
    else
        Result := FTable.ColWidths[index];
end;

procedure TPropertiesEdit.SetColWidth(index: integer; Value: integer);
begin
    if not index in [0,1] then exit;
    FTable.ColWidths[index] := Value;
end;

procedure TPropertiesEdit.SetEnabled(Value: boolean);
begin
    FEnabled := Value;
    if Visible then Paint;
end;

function TPropertiesEdit.GetNumProps(ObjIndex: integer) : integer;
var
    Obj : PObj;
begin
    Obj := FObjects.Items[ObjIndex];
    Result := Obj^.Props.Count;
end;

procedure TPropertiesEdit.FTableSelect(Sender: TObject; Col,Row: longint;
                                       var CanSelect: boolean);
begin
    CanSelect := True;
    if Row>0 then FPropertyIndex := PropertyIndexOf(FTable.Cells[0,Row]);
    if Visible then Paint;
end;

procedure TPropertiesEdit.FObjectChange(Sender: TObject);
begin
    FObjectIndex := FObjectNames.ItemIndex;
    if Visible then Paint;
end;

procedure TPropertiesEdit.FPropPullDownValuesChange(Sender: TObject);
var
    Obj     : PObj;
    Prop    : PProperty;
    changed : boolean;
begin
    Obj := FObjects.Items[FObjectIndex];
    Prop := Obj^.Props.Items[FPropertyIndex];
    with FPropPullDown do
        if VarType(Prop^.Value)<>varBoolean then begin
            changed := (Prop^.Value<>Items[ItemIndex]);
            VarCast(Prop^.Value,Items[ItemIndex],VarType(Prop^.Value)); end
        else if Items[ItemIndex]='True' then begin
            changed := not Prop^.Value;
            Prop^.Value := True; end
        else begin
            changed := Prop^.Value;
            Prop^.Value := False;
        end;
    if Visible then Paint;
    if Assigned(FOnPropertyChange) then FOnPropertyChange(Sender);
    if changed and Assigned(FOnPropertyValueChange) then
        FOnPropertyValueChange(Sender);
    FAfterEdit(Sender);
end;

procedure TPropertiesEdit.FPropEditChange(Sender: TObject);
var
    Obj     : PObj;
    Prop    : PProperty;
    changed : boolean;
begin
    changed := False;
    if (FObjectIndex<0) or (FPropertyIndex<0) then exit;
    Obj := FObjects.Items[FObjectIndex];
    Prop := Obj^.Props.Items[FPropertyIndex];
    case VarType(Prop^.Value) of
        varInteger,varSmallInt,varByte: begin
            changed := (Prop^.Value<>FPropIntegerEdit.Value);
            Prop^.Value := FPropIntegerEdit.Value;
        end;
        varSingle,varDouble,varCurrency: begin
            changed := (Prop^.Value<>FPropFloatEdit.Value);
            Prop^.Value := FPropFloatEdit.Value;
        end;
        varBoolean: begin
            if UpperCase(Trim(FPropStringEdit.Text))='TRUE' then begin
                changed := not Prop^.Value;
                Prop^.Value := True; end
            else if UpperCase(Trim(FPropStringEdit.Text))='FALSE' then begin
                changed := Prop^.Value;
                Prop^.Value := False;
            end;
        end;
    else
        changed := (VarToStr(Prop^.Value)<>FPropStringEdit.Text);
        Prop^.Value := FPropStringEdit.Text;
    end;
    if Visible then Paint;
    if Assigned(FOnPropertyChange) then FOnPropertyChange(Sender);
    if changed and Assigned(FOnPropertyValueChange) then
        FOnPropertyValueChange(Sender);
end;

procedure TPropertiesEdit.FPropEditKeyPress(Sender: TObject; var Key: char);
var
    Obj  : PObj;
    Prop : PProperty;
    tmp  : variant;
begin
    if Key=chr(13) then begin
        Obj := FObjects.Items[FObjectIndex];
        Prop := Obj^.Props.Items[FPropertyIndex];
        if Sender is TIntegerEdit then
            tmp := FPropIntegerEdit.Value
        else if Sender is TFloatEdit then
            tmp := FPropFloatEdit.Value
        else if VarType(Prop^.Value)=varBoolean then begin
            if UpperCase(FPropStringEdit.Text)='TRUE' then
                tmp := True
            else if UpperCase(FPropStringEdit.Text)='FALSE' then
                tmp := False; end
        else
            tmp := Prop^.Value;
        if (not VarIsEmpty(tmp)) and (tmp<>Prop^.Value) then
            FPropEditChange(Sender)
        else if Visible then 
            Paint;
    end;
end;

procedure TPropertiesEdit.FOKBtnClick(Sender: TObject);
var
    Obj  : PObj;
    Prop : PProperty;
begin
    if (FObjectIndex<0) or (FObjectIndex>=FNumObjects) then exit;
    Obj := FObjects.Items[FObjectIndex];
    if (FPropertyIndex<0) or (FPropertyIndex>=Obj^.Props.Count) then exit;
    Prop := Obj^.Props.Items[FPropertyIndex];
    if Prop^.EditMode in [emUserEdit,emUserEditBrowse,emUserBrowse] then begin
        case VarType(Prop^.Value) of
            varInteger,varSmallInt,varByte: Prop^.Value := FPropIntegerEdit.Value;
            varSingle,varDouble,varCurrency: Prop^.Value := FPropFloatEdit.Value;
            varBoolean: begin
                if UpperCase(FPropStringEdit.Text)='TRUE' then
                    Prop^.Value := True
                else if UpperCase(FPropStringEdit.Text)='FALSE' then
                    Prop^.Value := False;
            end;
        else
            Prop^.Value := FPropStringEdit.Text;
        end;
        FPropEditChange(self);
    end;
    FAfterEdit(Sender);
end;

procedure TPropertiesEdit.FCancelBtnClick(Sender: TObject);
var
    Obj  : PObj;
    Prop : PProperty;
begin
    if (FObjectIndex<0) or (FObjectIndex>=FNumObjects) then exit;
    Obj := FObjects.Items[FObjectIndex];
    if (FPropertyIndex<0) or (FPropertyIndex>=Obj^.Props.Count) then exit;
    Prop := Obj^.Props.Items[FPropertyIndex];
    if Prop^.EditMode in [emUserEdit,emUserEditBrowse,emUserBrowse] then begin
        case VarType(Prop^.Value) of
            varInteger,varSmallInt,varByte: FPropIntegerEdit.Value := Prop^.Value;
            varSingle,varDouble,varCurrency: FPropFloatEdit.Value := Prop^.Value;
            varBoolean: begin
                if Prop^.Value=True then
                    FPropStringEdit.Text := 'True'
                else
                    FPropStringEdit.Text := 'False';
            end;
        else
            FPropStringEdit.Text := Prop^.Value;
        end;
    end;
    FAfterEdit(Sender);
end;

procedure TPropertiesEdit.FBrowseBtnClick(Sender: TObject);
var
    Obj         : PObj;
    Prop        : PProperty;
    UserBrowser : TUserBrowser;
    changed     : boolean;
begin
    changed := False;
    if (FObjectIndex<0) or (FObjectIndex>=FNumObjects) then exit;
    Obj := FObjects.Items[FObjectIndex];
    if (FPropertyIndex<0) or (FPropertyIndex>=Obj^.Props.Count) then exit;
    Prop := Obj^.Props.Items[FPropertyIndex];
    case Prop^.EditMode of
         emUserBrowse,emUserEditBrowse: begin
            UserBrowser := Prop^.UserBrowser;
            if Assigned(UserBrowser) then begin
                changed := UserBrowser(Prop^.Value);
            end;
         end;
         emFileBrowse: begin
            FOpenDialog.Filter := Prop^.Filter;
            if FOpenDialog.Execute then begin
                Prop^.Value := FOpenDialog.Filename;
                if Visible then Paint;
                changed := True;
            end;
         end;
         emFontBrowse: begin
            FFontDialog.Font.Name := Prop^.Value;
            if FFontDialog.Execute then begin
                Prop^.Value := FFontDialog.Font.Name;
                if Visible then Paint;
                changed := True;
            end;
         end;
         emColorBrowse: begin
            FColorDialog.Color := Prop^.Value;
            if FColorDialog.Execute then begin
                Prop^.Value := FColorDialog.Color;
                if Visible then Paint;
                changed := True;
            end;
         end;
    end;
    if Assigned(FOnPropertyChange) then FOnPropertyChange(Sender);
    if changed and Assigned(FOnPropertyValueChange) then
        FOnPropertyValueChange(Sender);
    FAfterEdit(Sender);
end;

function TPropertiesEdit.GetPropertyName(index: integer): string;
var
    Obj  : PObj;
    Prop : PProperty;
begin
    if (FObjectIndex<0) or (FObjectIndex>=FObjects.Count) then exit;
    Obj := FObjects.Items[FObjectIndex];
    if (index>=Obj^.Props.Count) or (index<0) then exit;
    Prop := Obj^.Props.Items[index];
    Result := Prop^.Name;
end;

procedure TPropertiesEdit.SetPropertyName(index: integer; PropName: string);
var
    Obj  : PObj;
    Prop : PProperty;
begin
    if (FObjectIndex<0) or (FObjectIndex>=FObjects.Count) then exit;
    Obj := FObjects.Items[FObjectIndex];
    if (index>=Obj^.Props.Count) or (index<0) then exit;
    Prop := Obj^.Props.Items[index];
    Prop^.Name := PropName;
end;

function TPropertiesEdit.GetPropertyFilter(index: integer): string;
var
    Obj  : PObj;
    Prop : PProperty;
begin
    if (FObjectIndex<0) or (FObjectIndex>=FObjects.Count) then exit;
    Obj := FObjects.Items[FObjectIndex];
    if (index>=Obj^.Props.Count) or (index<0) then exit;
    Prop := Obj^.Props.Items[index];
    Result := Prop^.Filter;
end;

procedure TPropertiesEdit.SetPropertyFilter(index: integer; Filter: string);
var
    Obj  : PObj;
    Prop : PProperty;
begin
    if (FObjectIndex<0) or (FObjectIndex>=FObjects.Count) then exit;
    Obj := FObjects.Items[FObjectIndex];
    if (index>=Obj^.Props.Count) or (index<0) then exit;
    Prop := Obj^.Props.Items[index];
    Prop^.Filter := Filter;
end;

function TPropertiesEdit.GetValueSet(index: integer): variant;
var
    Obj  : PObj;
    Prop : PProperty;
begin
    if (FObjectIndex<0) or (FObjectIndex>=FObjects.Count) then exit;
    Obj := FObjects.Items[FObjectIndex];
    if (index>=Obj^.Props.Count) or (index<0) then exit;
    Prop := Obj^.Props.Items[index];
    Result := Prop^.ValueSet;
end;

procedure TPropertiesEdit.SetValueSet(index: integer; ValueSet: variant);
var
    Obj   : PObj;
    Prop  : PProperty;
    lo,hi : integer;
    found : boolean;
begin
    if (FObjectIndex<0) or (FObjectIndex>=FObjects.Count) then exit;
    Obj := FObjects.Items[FObjectIndex];
    if (index>=Obj^.Props.Count) or (index<0) then exit;
    Prop := Obj^.Props.Items[index];
    Prop^.ValueSet := ValueSet;
    if VarIsArray(ValueSet) then begin
        lo := VarArrayLowBound(ValueSet,1);
        hi := VarArrayHighBound(ValueSet,1);
        found := False;
        while (lo<=hi) and (not found) do begin
            found := (ValueSet[lo]=Prop^.Value);
            inc(lo);
        end;
        if not found then Prop^.Value := ValueSet[0];
    end;
end;

function TPropertiesEdit.GetEditMode(index: integer): integer;
var
    Obj  : PObj;
    Prop : PProperty;
begin
    if (FObjectIndex<0) or (FObjectIndex>=FObjects.Count) then
        Result := -1
    else begin
        Obj := FObjects.Items[FObjectIndex];
        if (index>=Obj^.Props.Count) or (index<0) then
            Result := -1
        else begin
            Prop := Obj^.Props.Items[index];
            Result := Prop^.EditMode;
        end;
    end;
end;

procedure TPropertiesEdit.SetEditMode(index,EditMode: integer);
var
    Obj  : PObj;
    Prop : PProperty;
begin
    if (FObjectIndex<0) or (FObjectIndex>=FObjects.Count) then exit;
    Obj := FObjects.Items[FObjectIndex];
    if (index>=Obj^.Props.Count) or (index<0) then exit;
    Prop := Obj^.Props.Items[index];
    Prop^.EditMode := EditMode;
end;

function TPropertiesEdit.GetPropertyBrowser(index: integer): TUserBrowser;
var
    Obj  : PObj;
    Prop : PProperty;
begin
    if (FObjectIndex<0) or (FObjectIndex>=FObjects.Count) then
        Result := nil
    else begin
        Obj := FObjects.Items[FObjectIndex];
        if (index>=Obj^.Props.Count) or (index<0) then
            Result := nil
        else begin
            Prop := Obj^.Props.Items[index];
            Result := Prop^.UserBrowser;
        end;
    end;
end;

procedure TPropertiesEdit.SetPropertyBrowser(index: integer; Browser: TUserBrowser);
var
    Obj  : PObj;
    Prop : PProperty;
begin
    if (FObjectIndex<0) or (FObjectIndex>=FObjects.Count) then exit;
    Obj := FObjects.Items[FObjectIndex];
    if (index>=Obj^.Props.Count) or (index<0) then exit;
    Prop := Obj^.Props.Items[index];
    Prop^.UserBrowser := Browser;
end;

function TPropertiesEdit.GetVisible(index: integer): boolean;
var
    Obj  : PObj;
    Prop : PProperty;
begin
    if (FObjectIndex<0) or (FObjectIndex>=FObjects.Count) then
        Result := False
    else begin
        Obj := FObjects.Items[FObjectIndex];
        if (index>=Obj^.Props.Count) or (index<0) then
            Result := False
        else begin
            Prop := Obj^.Props.Items[index];
            Result := Prop^.Visible and Prop^.DefVisible;
        end;
    end;
end;

procedure TPropertiesEdit.SetVisible(index: integer; Value: boolean);
var
    Obj  : PObj;
    Prop : PProperty;
begin
    if (FObjectIndex<0) or (FObjectIndex>=FObjects.Count) then exit;
    Obj := FObjects.Items[FObjectIndex];
    if (index>=Obj^.Props.Count) or (index<0) then exit;
    Prop := Obj^.Props.Items[index];
    Prop^.DefVisible := Value;
    if Visible then Paint;
end;

procedure TPropertiesEdit.SetSorted(Value: boolean);
var
    Obj : PObj;
    i   : integer;
begin
    if (not FSorted) and Value then begin
        for i := 0 to FObjects.Count-1 do begin
            Obj := FObjects.Items[i];
            Obj^.Props.Sort(CompareProp);
        end;
        FPropertyIndex := -1;
        if Visible then Paint;
    end;
    FSorted := Value;
end;

procedure TPropertiesEdit.SpinUp(Sender: TObject);
var
    i : integer;
begin
    i := FPropIntegerEdit.Value;
    FPropIntegerEdit.Value := i-1;
    FPropEditChange(Sender);
    FAfterEdit(Sender);
end;

procedure TPropertiesEdit.SpinDown(Sender: TObject);
var
    i : integer;
begin
    i := FPropIntegerEdit.Value;
    FPropIntegerEdit.Value := i+1;
    FPropEditChange(Sender);
    FAfterEdit(Sender);
end;

function TPropertiesEdit.ObjectIndexOf(Name: string): integer;
var
    i     : integer;
    found : boolean;
begin
    found := False;
    i := 0;
    while (i<FObjects.Count) and (not found) do begin
        found := (PObj(FObjects.Items[i])^.Name=Name);
        inc(i);
    end;
    if found then
        Result := i-1
    else
        Result := -1;
end;

function TPropertiesEdit.PropertyIndexOf(Name: string): integer;
var
    i     : integer;
    Obj   : PObj;
    found : boolean;
begin
    if (FObjectIndex<0) or (FObjectIndex>=FObjects.Count) then
        Result := -1
    else begin
        Obj := FObjects.Items[FObjectIndex];
        found := False;
        i := 0;
        while (i<Obj^.Props.Count) and (not found) do begin
            found := (PProperty(Obj^.Props.Items[i])^.Name=Name);
            inc(i);
        end;
        if found then
            Result := i-1
        else
            Result := -1;
    end;
end;

function TPropertiesEdit.GetNDigits(index: integer): byte;
var
    Obj  : PObj;
    Prop : PProperty;
begin
    if (FObjectIndex<0) or (FObjectIndex>=FObjects.Count) then
        Result := 0
    else begin
        Obj := FObjects.Items[FObjectIndex];
        if (index>=Obj^.Props.Count) or (index<0) then
            Result := 0
        else begin
            Prop := Obj^.Props.Items[index];
            Result := Prop^.NDigits;
        end;
    end;
end;

procedure TPropertiesEdit.SetNDigits(index: integer; Value: byte);
var
    Obj  : PObj;
    Prop : PProperty;
begin
    if (FObjectIndex<0) or (FObjectIndex>=FObjects.Count) then exit;
    Obj := FObjects.Items[FObjectIndex];
    if (index>=Obj^.Props.Count) or (index<0) then exit;
    Prop := Obj^.Props.Items[index];
    Prop^.NDigits := Value;
    if Visible then Paint;
end;

procedure TPropertiesEdit.SetAutoHide(Value: boolean);
begin
    if FAutoHide<>Value then begin
        FAutoHide := Value;
        if Visible then Paint;
    end;
end;

function TPropertiesEdit.GetDependOn(index: integer): string;
var
    Obj  : PObj;
    Prop : PProperty;
begin
    if (FObjectIndex<0) or (FObjectIndex>=FObjects.Count) then
        Result := ''
    else begin
        Obj := FObjects.Items[FObjectIndex];
        if (index>=Obj^.Props.Count) or (index<0) then
            Result := ''
        else begin
            Prop := Obj^.Props.Items[index];
            Result := Prop^.DependOn;
        end;
    end;
end;

procedure TPropertiesEdit.SetDependOn(index: integer; Value: string);
var
    Obj  : PObj;
    Prop : PProperty;
begin
    if (FObjectIndex<0) or (FObjectIndex>=FObjects.Count) then exit;
    Obj := FObjects.Items[FObjectIndex];
    if (index>=Obj^.Props.Count) or (index<0) then exit;
    Prop := Obj^.Props.Items[index];
    Prop^.DependOn := Value;
    if Visible then Paint;
end;

function TPropertiesEdit.GetDependValue(index: integer): variant;
var
    Obj  : PObj;
    Prop : PProperty;
begin
    if (FObjectIndex<0) or (FObjectIndex>=FObjects.Count) then
        VarClear(Result)
    else begin
        Obj := FObjects.Items[FObjectIndex];
        if (index>=Obj^.Props.Count) or (index<0) then
            VarClear(Result)
        else begin
            Prop := Obj^.Props.Items[index];
            Result := Prop^.DependValue;
        end;
    end;
end;

procedure TPropertiesEdit.SetDependValue(index: integer; Value: variant);
var
    Obj  : PObj;
    Prop : PProperty;
begin
    if (FObjectIndex<0) or (FObjectIndex>=FObjects.Count) then exit;
    Obj := FObjects.Items[FObjectIndex];
    if (index>=Obj^.Props.Count) or (index<0) then exit;
    Prop := Obj^.Props.Items[index];
    Prop^.DependValue := Value;
    if Visible then Paint;
end;

procedure TPropertiesEdit.ExchangeProperty(index1, index2: integer);
var
    Obj : PObj;
begin
    if (FObjectIndex<0) or (FObjectIndex>=FObjects.Count) then exit;
    Obj := FObjects.Items[FObjectIndex];
    if (index1>=0) and (index1<Obj^.Props.Count) and
       (index2>=0) and (index2<Obj^.Props.Count) then begin
        Obj^.Props.Exchange(index1,index2);
        if Visible then Paint;
    end;
end;

procedure TPropertiesEdit.ExchangeObject(index1, index2: integer);
begin
    if (index1>=0) and (index1<FObjects.Count) and
       (index2>=0) and (index2<FObjects.Count) then begin
        FObjects.Exchange(index1,index2);
        FObjectNames.Items.Exchange(index1,index2);
        if Visible then Paint;
    end;
end;

procedure TPropertiesEdit.Paint;
var
    i,j,k,hi,lo : integer;
    Obj         : PObj;
    Prop        : PProperty;
    tVisible    : boolean;
    Propname    : string;
begin
    inherited Paint;
    Propname := Propertyname[FPropertyIndex];
    for i := 0 to ComponentCount-1 do
        if not((Components[i] is TOpenDialog) or
               (Components[i] is TFontDialog) or
               (Components[i] is TColorDialog)) then
            TWinControl(Components[i]).Enabled := FEnabled;
    FObjectNames.Visible := FShowObjectList;
    if FObjectNames.Visible then
        FCancelBtn.Top := FObjectNames.Top+FObjectNames.Height
    else
        FCancelBtn.Top := 0;
    FOKBtn.Top := FCancelBtn.Top;
{Adjust Link button}
    with FBrowseBtn do begin
        Visible := False;
        SendToBack;
        SetBounds(self.ClientWidth-FOKBtn.Width,FOKBtn.Top,Width,Height);
    end;
{Adjust edit components for property value}
    with FPropStringEdit do begin
        SetBounds(FOKBtn.Left+FOKBtn.Width+2,FOKBtn.Top,
                  self.ClientWidth-Left-FBrowseBtn.Width,Height);
    end;
    with FPropFloatEdit do begin
        SetBounds(FPropStringEdit.Left,FPropStringEdit.Top,
                  FPropStringEdit.Width,Height);
    end;
    with FPropIntegerEdit do begin
        SetBounds(FPropStringEdit.Left,FPropStringEdit.Top,
                  FPropStringEdit.Width-FPropIntegerUpDown.Width,Height);
    end;
    with FPropIntegerUpDown do begin
        SetBounds(FPropIntegerEdit.Left+FPropIntegerEdit.Width,
                  FPropIntegerEdit.Top,15,FPropIntegerEdit.Height);
    end;
{Adjust pull down list}
    with FPropPullDown do begin
        Visible := False;
        SetBounds(FPropStringEdit.Left,FPropStringEdit.Top,
                  self.ClientWidth-FPropPullDown.Left,Height);
    end;
{Adjust table}
    with FTable do begin
        SetBounds(FObjectNames.Left,FPropStringEdit.Top+FPropStringEdit.Height,
                  self.ClientWidth,self.ClientHeight-Top);
    end;
{Fill in the blank}
    FObjectNames.ItemIndex := FObjectIndex;
    if (FObjectIndex<0) or (FObjectIndex>=FNumObjects) then begin
        for i := 0 to ComponentCount-1 do
            if not((Components[i] is TOpenDialog) or
                   (Components[i] is TFontDialog) or
                   (Components[i] is TColorDialog)) then
                TWinControl(Components[i]).Invalidate;
        exit;
    end;
    Obj := FObjects.Items[FObjectIndex];
    j := 1;
    for i := 1 to Obj^.Props.Count do with FTable do begin
        Prop := Obj^.Props.Items[i-1];
        tVisible := Prop^.DefVisible;
        k := PropertyIndexOf(Prop^.DependOn);
        if FAutoHide and (k>=0) then
            tVisible := tVisible and IsVisible[k] and
                        (VarToStr(Value[k])=VarToStr(Prop^.DependValue));
        Prop^.Visible := tVisible;
        if tVisible then begin
            Cells[0,j] := Prop^.Name;
            if Prop^.EditMode=emColorBrowse then begin
                Cells[1,j] := '[TColor]';
                inc(j);
                continue;
            end;
            if Prop^.EditMode=emUserBrowse then begin
                Cells[1,j] := '[....]';
                inc(j);
                continue;
            end;
            case VarType(Prop^.Value) of
                varBoolean :
                    if Prop^.Value=True then
                        Cells[1,j] := 'True'
                    else
                        Cells[1,j] := 'False';
                varSingle,varDouble,varCurrency :
                    if Prop^.EditMode<>emDropDownList then
                        Cells[1,j] := FloatToStrF(Prop^.Value,ffFixed,18,Prop^.NDigits)
                    else
                        Cells[1,j] := VarToStr(Prop^.Value);
                else
                    Cells[1,j] := VarToStr(Prop^.Value);
            end;
            if VarIsArray(Prop^.Value) or VarIsEmpty(Prop^.Value) then
                Cells[1,j] := '[...]';
            inc(j);
        end;
    end;
    if (j<2) and (FTable.FixedRows=1) then FTable.FixedRows := 0;    
    if FTable.RowCount<>j then FTable.RowCount := j;
    if (j>1) and (FTable.FixedRows=0) then FTable.FixedRows := 1;
    if (j>1) and (FPropertyIndex<0) then FPropertyIndex := 0;
    if (FPropertyIndex>=Obj^.Props.Count) or (FPropertyIndex<0) then begin
        FPropPullDown.SendToBack;
        FPropPullDown.Visible := False;
        FPropStringEdit.SendToBack;
        FPropStringEdit.Visible := False;
        FPropFloatEdit.SendToBack;
        FPropFloatEdit.Visible := False;
        FPropIntegerEdit.SendToBack;
        FPropIntegerEdit.Visible := False;
        FPropIntegerUpDown.SendToBack;
        FPropIntegerUpDown.Visible := False;
        FBrowseBtn.SendToBack;
        FBrowseBtn.Visible := False;
        FObjectNames.Enabled := FEnabled;
        FTable.Enabled := FEnabled;
        FTable.Invalidate;
        exit; end
    else begin
        j := 1;
        while (j<FTable.RowCount) and
              (FTable.Cells[0,j]<>Propname) do inc(j);
        FTable.Selection := TGridRect(Rect(0,j,1,j));
    end;
    Prop := Obj^.Props.Items[FPropertyIndex];
    case Prop^.EditMode of
        emUserEdit: begin
            case VarType(Prop^.Value) of
                varSmallInt,varInteger,varByte: begin
                    with FPropIntegerEdit do begin
                        BringToFront;
                        Visible := True;
                        ReadOnly := False;
                        ClipMax := False;
                        ClipMin := False;
                        Max := Prop^.MaxValue;
                        Min := Prop^.MinValue;
                        MaxLength := Prop^.MaxLength;
                        Value := Prop^.Value;
                        ClipMax := True;
                        ClipMin := True;
                    end;
                    with FPropIntegerUpDown do begin
                        BringToFront;
                        Visible := True;
                    end;
                    FBrowseBtn.RePaint;
                    FPropIntegerEdit.RePaint;
                    FPropIntegerUpDown.RePaint;
                    FPropPullDown.RePaint;
                end;
                varSingle,varDouble,varCurrency : begin
                    with FPropFloatEdit do begin
                        BringToFront;
                        Visible := True;
                        ReadOnly := False;
                        ClipMax := False;
                        ClipMin := False;
                        Max := Prop^.MaxValue;
                        Min := Prop^.MinValue;
                        MaxLength := Prop^.MaxLength;
                        Value := Prop^.Value;
                        Digits := Prop^.NDigits;
                        ClipMax := True;
                        ClipMin := True;
                    end;
                end;
            else with FPropStringEdit do begin
                    BringToFront;
                    Visible := True;
                    ReadOnly := False;
                    MaxLength := Prop^.MaxLength;
                    if VarType(Prop^.Value)=varBoolean then
                        if Prop^.Value=True then
                            Text := 'True'
                        else
                            Text := 'False'
                    else
                        Text := VarToStr(Prop^.Value);
            end; end;
            FBrowseBtn.BringToFront;
            FBrowseBtn.Visible := True;
            FBrowseBtn.Enabled := False;
            FOKBtn.Enabled := FEnabled;
            FCancelBtn.Enabled := FEnabled;
        end;
        emDropDownList: begin
            if VarIsArray(Prop^.ValueSet) then begin
                hi := VarArrayHighBound(Prop^.ValueSet,1);
                lo := VarArrayLowBound(Prop^.ValueSet,1); end
            else begin
                hi := 0;
                lo := 1;
            end;
            FBrowseBtn.Visible := False;
            FPropPullDown.BringToFront;
            FPropPullDown.Visible := True;
            FPropPullDown.Clear;
            for i := lo to hi do begin
                if VarType(Prop^.Value)<>varBoolean then
                    FPropPullDown.Items.Add(VarToStr(Prop^.ValueSet[i]))
                else if Prop^.ValueSet[i]=True then
                    FPropPullDown.Items.Add('True')
                else
                    FPropPullDown.Items.Add('False');
                if VarToStr(Prop^.ValueSet[i])=VarToStr(Prop^.Value) then
                    FPropPullDown.ItemIndex := i-lo;
            end;
            FOKBtn.Enabled := False;
            FCancelBtn.Enabled := False;
        end;
        emUserBrowse: begin
            FPropStringEdit.BringToFront;
            FPropStringEdit.Visible := True;
            case VarType(Prop^.Value) of
                varBoolean :
                    if Prop^.Value=True then
                        FPropStringEdit.Text := 'True'
                    else
                        FPropStringEdit.Text := 'False';
                else
                    FPropStringEdit.Text := '[...]';
            end;
            FPropStringEdit.ReadOnly := True;
            FBrowseBtn.Visible := True;
            FBrowseBtn.Enabled := FEnabled;
            FOKBtn.Enabled := False;
            FCancelBtn.Enabled := False;
        end;
        emUserEditBrowse: begin
            case VarType(Prop^.Value) of
                varSmallInt,varInteger,varByte: begin
                    with FPropIntegerEdit do begin
                        BringToFront;
                        Visible := True;
                        ReadOnly := False;
                        ClipMax := False;
                        ClipMin := False;
                        Max := Prop^.MaxValue;
                        Min := Prop^.MinValue;
                        MaxLength := Prop^.MaxLength;
                        Value := Prop^.Value;
                        ClipMax := True;
                        ClipMin := True;
                    end;
                    with FPropIntegerUpDown do begin
                        BringToFront;
                        Visible := True;
                    end;
                    FBrowseBtn.RePaint;
                    FPropIntegerEdit.RePaint;
                    FPropIntegerUpDown.RePaint;
                    FPropPullDown.RePaint;
                end;
                varSingle,varDouble,varCurrency : begin
                    with FPropFloatEdit do begin
                        BringToFront;
                        Visible := True;
                        ReadOnly := False;
                        ClipMax := False;
                        ClipMin := False;
                        Max := Prop^.MaxValue;
                        Min := Prop^.MinValue;
                        MaxLength := Prop^.MaxLength;
                        Value := Prop^.Value;
                        Digits := Prop^.NDigits;
                        ClipMax := True;
                        ClipMin := True;
                    end;
                end;
            else with FPropStringEdit do begin
                    BringToFront;
                    Visible := True;
                    ReadOnly := False;
                    MaxLength := Prop^.MaxLength;
                    if VarType(Prop^.Value)=varBoolean then
                        if Prop^.Value=True then
                            Text := 'True'
                        else
                            Text := 'False'
                    else if VarIsArray(Prop^.Value) then
                        Text := '[...]'
                    else
                        Text := VarToStr(Prop^.Value);
            end; end;
            FBrowseBtn.BringToFront;
            FBrowseBtn.Visible := True;
            FBrowseBtn.Enabled := FEnabled;
            FOKBtn.Enabled := FEnabled;
            FCancelBtn.Enabled := FEnabled;
        end;
        emColorBrowse: begin
            FPropStringEdit.BringToFront;
            FPropStringEdit.Visible := True;
            FPropStringEdit.ReadOnly := True;
            FPropStringEdit.Text := '[TColor]';
            FBrowseBtn.BringToFront;
            FBrowseBtn.Visible := True;
            FBrowseBtn.Enabled := FEnabled;
            FOKBtn.Enabled := False;
            FCancelBtn.Enabled := False;
        end;
        emFileBrowse,emFontBrowse: begin
            FPropStringEdit.BringToFront;
            FPropStringEdit.Visible := True;
            case VarType(Prop^.Value) of
                varBoolean : begin
                    if Prop^.Value=True then
                        FPropStringEdit.Text := 'True'
                    else
                        FPropStringEdit.Text := 'False';
                    FPropStringEdit.ReadOnly := True;
                end;
                varArray or varVariant: begin
                    FPropStringEdit.Text := '[...]';
                    FPropStringEdit.ReadOnly := True;
                end;
            else
                FPropStringEdit.Text := VarToStr(Prop^.Value);
            end;
            FBrowseBtn.BringToFront;
            FBrowseBtn.Visible := True;
            FBrowseBtn.Enabled := FEnabled;
            FOKBtn.Enabled := False;
            FCancelBtn.Enabled := False;
        end;
    end;
end;

{Implementation of component editor}
procedure TCompEditor.ExecuteVerb(Index: integer);
var
    About : TAboutBox;
    QuickRef : TQuickRef;
begin
    case Index of
        0 : begin
            About := TAboutBox.Create(Application);
            About.Execute;
            About.Free;
        end;
        1 : begin
            QuickRef := TQuickRef.Create(Application);
            QuickRef.Execute;
            QuickRef.Free;
        end;
    end;
end;

function TCompEditor.GetVerb(Index: integer): string;
begin
    case Index of
        0 : Result := 'About ...';
        1 : Result := 'Quick Reference';
    end;
end;

function TCompEditor.GetVerbCount: integer;
begin
    Result := 2;
end;

procedure Register;
begin
    RegisterComponents('William', [TPropertiesEdit]);
    RegisterComponentEditor(TPropertiesEdit,TCompEditor);
end;

end.
