unit TPEREF;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ComCtrls;

type
  TQuickRef = class(TForm)
    RichEdit1: TRichEdit;
    Button1: TButton;
    procedure Button1Click(Sender: TObject);
    procedure FormCreate(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    procedure Execute;
  end;

implementation

{$R *.DFM}

const
    Props : array[0..19] of string =
        ('AutoHide: boolean',
         'Browser[PropertyIndex]: TUserBrowser',
         'ColWidths: integer',
         'DependOn[PropertyIndex]: string',
         'DependValue[PropertyIndex]: variant',
         'EditMode[PropertyIndex]: integer',
         'Filter[PropertyIndex]: string',
         'IsVisible[PropertyIndex]: boolean',
         'MaxLength[PropertyIndex]: word',
         'MaxValue[PropertyIndex]: variant',
         'MinValue[PropertyIndex]: variant',
         'NDigits[PropertyIndex]: byte',
         'NumObjs: integer',
         'NumProps[ObjectIndex: integer]: integer',
         'ObjectIndex: integer',
         'PropertyIndex: integer',
         'PropertyName[PropertyIdnex]: string',
         'ShowObjectList: boolean',
         'Sorted: boolean',
         'Value[PropertyIndex: integer]: variant');
    PropsDesc : array[0..19] of string =
        ('- return/set the auto-hide feature',
         '- return/set the user browser for the indexed property',
         '- return/set the column width in the property table',
         '- return/set the property name on which the property depends',
         '- return/set the value to check for dependence',
         '- return/set the editing mode for the indexed property',
         '- return/set the file filter used for the indexed property',
         '- return/set the visible property of the indexed property',
         '- return/set the max length for inputing indexed property, 0 for any',
         '- return/set the max value for the indexed property',
         '- return/set the min value for the indexed property',
         '- return/set the number of decimal places for floating point property',
         '- return the number of objects in the editor',
         '- return the number of properties of object specified by ObjectIndex',
         '- return/set the index of active object',
         '- return/set the index of active property',
         '- return/set the name of the indexed property',
         '- return/set the visible property of the object drop down list',
         '- return/set the sorting status of the property editor',
         '- return the value of the indexed property of object specified'+
             ' by ObjectIndex');
    Methods : array[0..13] of string =
        ('function AddObject(Name: string) : integer;',
         'function AddPropertyConst(Name: string; EditMode: integer;'+chr(10)+
         '    const ValueSet: array of variant; Value: variant) : integer;',
         'function AddPropertyVar(Name: string; EditMode: integer;'+chr(10)+
         '    ValueSet: variant; Value: variant) : integer;',
         'procedure Clear;',
         'construction Create(AOwner: TComponent);',
         'procedure DeleteObject(index: integer);',
         'procedure DeleteProperty(index: integer);',
         'procedure ExchangeObject(index1, index2: integer)',
         'procedure ExchangeProperty(index1, index2: integer)',
         'function InsertObject(index: integer; Name: string) : integer;',
         'function InsertPropertyConst(index: integer; Name: string;'+chr(10)+
         '    EditMode: integer; const ValueSet: array of variant;'+chr(10)+
         '    Value: variant) : integer;',
         'function InsertPropertyVar(index: iteger; Name: string;'+chr(10)+
         '    EditMode: integer; ValueSet: variant;'+chr(10)+
         '    Value: variant) : integer;',
         'function ObjectIndexOf(Name: string): integer;',
         'function PropertyIndexOf(Name: string): integer;');
    MethodsDesc : array[0..13] of string =
        ('- add an object to the editor and return the object index',
         '- add a property to the active object with const value set and return the property index',
         '- add a property to the active object with variable value set and return the property index',
         '- remove all objects/properties',
         '- create an editor',
         '- remove the object and its properties with object index index',
         '- remove the indexed property from the active object',
         '- exchange the indexed obejcts',
         '- exchange the indexed properties from the active object',
         '- insert an object to the editor at pos index',
         '- insert a property to the active object with const value set at pos index',
         '- insert a property to the active object with variable value set at pos index',
         '- return the index of the named object, -1 if not found',
         '- return the index of the named property, -1 if not found');
    Events : array[0..3] of string =
        ('OnEnter: TNotifyEvent;',
         'OnExit: TNotifyEvent;',
         'OnPropertyChange: TNotifyEvent;',
         'OnPropertyValueChange: TNotifyEvent;');
    EventsDesc : array[0..3] of string =
        ('- occurs when the user switch to the editor',
         '- occurs when the user switch away from the editor',
         '- occurs when any property changes',
         '- occurs when any property''s value changed');

procedure TQuickRef.Execute;
begin
    Left := (Screen.Width-Width) div 2;
    Top  := (Screen.Height-Height) div 2;
    ShowModal;
end;

procedure TQuickRef.Button1Click(Sender: TObject);
begin
    Hide;
end;

procedure TQuickRef.FormCreate(Sender: TObject);
var
    i : integer;
begin
    with RichEdit1 do begin
        Clear;
        SelAttributes.Size := 8;
        SelAttributes.Color := clBlue;
        SelAttributes.Style := [fsBold];
        Lines.Add('Key Properties');
        Lines.Add('');
        SelAttributes.Color := clBlack;
        for i := Low(Props) to High(Props) do begin
            SelAttributes.Style := [fsBold];
            Lines.Add(Props[i]);
            SelAttributes.Style := [];
            Lines.Add(PropsDesc[i]);
            Lines.Add('');
        end;
        SelAttributes.Color := clBlue;
        SelAttributes.Style := [fsBold];
        Lines.Add('Methods');
        Lines.Add('');
        SelAttributes.Color := clBlack;
        for i := Low(Methods) to High(Methods) do begin
            SelAttributes.Style := [fsBold];
            Lines.Add(Methods[i]);
            SelAttributes.Style := [];
            Lines.Add(MethodsDesc[i]);
            Lines.Add('');
        end;
        SelAttributes.Color := clBlue;
        SelAttributes.Style := [fsBold];
        Lines.Add('Events');
        Lines.Add('');
        SelAttributes.Color := clBlack;
        for i := Low(Events) to High(Events) do begin
            SelAttributes.Style := [fsBold];
            Lines.Add(Events[i]);
            SelAttributes.Style := [];
            Lines.Add(EventsDesc[i]);
            Lines.Add('');
        end;
        SelStart := 0;
        SelLength := 0;
    end;
end;

end.
